;;;; This file is part of XMLTools
;;;; Copyright (C) 2004 Sergey Poznyakoff
;;;;
;;;; GAMLET is A Markup LanguagE Translator
;;;;
;;;; XMLTools is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; XMLTools is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with XMLTools; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;

(define-module (xmltools gamlet))

(use-modules (ice-9 rdelim))

(define gamlet-pp-open #\[)
(define gamlet-pp-close #\])

(define output-file-list '())
(define output-file-index 0)

(define temp-dir #f)
(define remove-working-dir #f)
(define working-dir #f)
(define output-dir #f)
(define postprocess-phase #f)
(define created-directories '())

(define-macro (forbid-postprocess-phase symbol)
  `(if postprocess-phase
       (throw 'misc-error ,symbol
	      "~A cannot be used in postprocessing phase" (list ,symbol) #f)))
	   
(define-public (gamlet:output-file-count)
  (length output-file-list))

(define-public (gamlet:output-file-index)
  output-file-index)

(define-public (gamlet:output-file-name index)
  (list-ref output-file-list index))

(define-public (gamlet:mkdir-p name)
  (catch 'system-error
	 (lambda ()
	   (mkdir name))
	 (lambda args
	   (let ((error-code (car (list-ref args 4))))
	     (cond
	      ((= error-code ENOENT)
	       (let ((p (string-rindex name #\/)))
		 (if (not p)
		     (apply throw args))
		 (gamlet:mkdir-p (substring name 0 p))
		 (gamlet:mkdir-p name)))
	      ((= error-code EEXIST));; Exit
	      (else
	       (apply throw args)))))))

(define (rmdir-r name)
  (letrec ((clear-cwd
	    (lambda ()
	      (let ((dir (opendir ".")))
		(do ((item (readdir dir) (readdir dir)))
		    ((eof-object? item))
		  (cond
		   ((or (string=? item ".") (string=? item ".."))) ; continue
		   ((eq? (stat:type (lstat item)) 'directory)
		    (chdir item)
		    (clear-cwd)
		    (chdir "..")
		    (rmdir item))
		   (else
		    (delete-file item))))
		(closedir dir)))))
    (let ((pwd (getcwd)))
      (catch #t
	     (lambda ()
	       (chdir name)
	       (clear-cwd)
	       (chdir pwd)
	       (rmdir name))
	     (lambda args
	       (chdir pwd)
	       (apply throw args))))))

(define (make-file-name dir name)
  (let ((full-file-name (cond
			 ((not dir)
			  name)
			 ((char=? (string-ref dir (-
						   (string-length dir) 1)) #\/)
			  (string-append dir name))
			 (else
			  (string-append dir "/" name)))))
    (cond
     ((string-rindex full-file-name #\/) =>
      (lambda (p)
	(let ((dir-name (substring full-file-name 0 p)))
	  (if (not (member dir-name created-directories))
	      (begin
		(gamlet:mkdir-p dir-name)
		(set! created-directories
		      (cons dir-name created-directories))))))))
    full-file-name))

(define (get-working-dir)
  (if (not working-dir)
      (begin
	(set! working-dir (make-file-name
			   (or temp-dir
			       (getenv "TMPDIR")
			       (getenv "TEMPDIR")
			       "/tmp")
			   (string-append
			    "gamlet"
			    (number->string (getpid)))))
	(set! remove-working-dir #t)))
  (if (not (access? working-dir F_OK))
      (gamlet:mkdir-p working-dir))
  working-dir)

(define-public (gamlet:change-pp-delimiter odelim cdelim)
  (cond
   ((not (char? odelim))
    (throw 'misc-error 'gamlet:change-pp-delimiter 
	   "Opening delimiter is not a character: ~S"
	     (list odelim) #f))
   ((not (char? cdelim))
    (throw 'misc-error 'gamlet:change-pp-delimiter 
	   "Closing delimiter is not a character: ~S"
	     (list cdelim) #f))
   (else
    (set! gamlet-pp-open odelim)
    (set! gamlet-pp-close cdelim))))

(define-public (gamlet:get-pp-delimiter)
  (cons gamlet-pp-open gamlet-pp-close))

(define gamlet-handler-list '())

(define-public (gamlet:set-handler name handler)
  (set! gamlet-handler-list (cons
			     (cons name handler)
			     gamlet-handler-list)))

(define (process-line str)
  (cond
   ((string-null? str)
    #f)
   ((string-index str gamlet-pp-open) =>
    (lambda (start)
      (let ((stop (string-index str gamlet-pp-close start)))
	(cond
	 (stop
	  (let ((elem (assoc (substring str (1+ start) stop)
			     gamlet-handler-list)))
	    (cond
	     (elem
	      (display (substring str 0 start))
	      ((cdr elem))
	      (process-line (substring str (1+ stop))))
	     (else
	      (display (substring str 0 (1+ start)))
	      (process-line (substring str (1+ start)))))))
	 (else
	  (display (substring str 0 (1+ start)))
	  (process-line (substring str (1+ start))))))))
   (else
    (display str))))
		 
(define (postprocess file-name)
  (with-input-from-file
      (make-file-name (get-working-dir) file-name)
    (lambda ()
      (let ((out-file-name (make-file-name output-dir file-name)))
	(if (access? out-file-name F_OK)
	    (delete-file out-file-name))
	(with-output-to-file
	    out-file-name
	  (lambda ()
	    (do ((line (read-line) (read-line)))
		((eof-object? line))
	      (process-line line)
	      (newline))))))))
    
(define (flush-files)
  (set! output-file-list (reverse output-file-list))
  (set! output-file-index 0)
  (for-each
   (lambda (file-name)
     (postprocess file-name)
     (set! output-file-index (1+ output-file-index)))
   output-file-list))
  
(define-public (gamlet:init)
  (set! postprocess-phase #f)
  (set! temp-dir #f)
  (set! remove-working-dir #f)
  (set! working-dir #f)
  (set! output-dir #f)
  (set! created-directories '()))

(define-public (gamlet:set-working-dir name)
  (if working-dir
      (throw 'misc-error 'gamlet:set-working-dir
	     "working directory already set to ~S" (list working-dir) #f))
  (set! working-dir name)
  (set! remove-working-dir #f))

(define-public (gamlet:set-output-dir name)
  (if output-dir
      (throw 'misc-error 'gamlet:set-output-dir
	     "output directory already set to ~S" (list output-dir) #f))
  (set! output-dir name))

(define-public (gamlet:open-output-file name)
  (forbid-postprocess-phase 'gamlet:open-output-file)
  (let ((port (make-file-name (get-working-dir) name)))
    (set! output-file-list (cons name output-file-list))
    port))

(define-public (gamlet:with-output-to-file name thunk)
  (forbid-postprocess-phase 'gamlet:with-output-to-file)
  (set! output-file-list (cons name output-file-list))
  (with-output-to-file
      (make-file-name (get-working-dir) name)
    thunk))

(define-public (gamlet:flush)
  (forbid-postprocess-phase 'gamlet:flush)
  (set! postprocess-phase #t)
  (flush-files)
  (if remove-working-dir
      (rmdir-r working-dir)))

(define-public (gamlet:current-output-file-name)
  (if postprocess-phase
      (gamlet:output-file-name (gamlet:output-file-index))
      (car output-file-list)))

;;;; End of gamlet.scm